from __future__ import annotations

# standard libraries
import gettext
import typing

# third party libraries
import numpy
import scipy.stats

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift.model import DataItem
from nion.ui import Declarative
from nion.utils import Converter
from nion.utils import Model

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController

_ = gettext.gettext


class GenerateDataDialog(Declarative.WindowHandler):

    def __init__(self, document_controller: DocumentController.DocumentController):
        super().__init__()

        self.__document_controller = document_controller

        self.title_model: Model.PropertyModel[str] = Model.PropertyModel(_("Generated Data"))
        self.data_type_model: Model.PropertyModel[int] = Model.PropertyModel(0)
        self.is_sequence_model: Model.PropertyModel[int] = Model.PropertyModel(0)
        self.collection_rank_model: Model.PropertyModel[int] = Model.PropertyModel(0)
        self.datum_rank_model: Model.PropertyModel[int] = Model.PropertyModel(1)

        self.sequence_size_model: Model.PropertyModel[int] = Model.PropertyModel(16)

        self.line_size_model: Model.PropertyModel[int] = Model.PropertyModel(512)

        self.scan_width_model: Model.PropertyModel[int] = Model.PropertyModel(256)
        self.scan_height_model: Model.PropertyModel[int] = Model.PropertyModel(256)

        self.spectrum_size_model: Model.PropertyModel[int] = Model.PropertyModel(1024)

        self.image_width_model: Model.PropertyModel[int] = Model.PropertyModel(1024)
        self.image_height_model: Model.PropertyModel[int] = Model.PropertyModel(1024)

        self.array_width_model: Model.PropertyModel[int] = Model.PropertyModel(1024)
        self.array_height_model: Model.PropertyModel[int] = Model.PropertyModel(256)

        self.int_converter = Converter.IntegerToStringConverter()

        u = Declarative.DeclarativeUI()

        title_row = u.create_row(u.create_label(text=_("Title")),
                                 u.create_line_edit(text="@binding(title_model.value)"), u.create_stretch(), spacing=8)

        data_type_row = u.create_row(u.create_label(text=_("Data Type")), u.create_combo_box(items=["Float (32-bit)"],
                                                                                             current_index="@binding(data_type_model.value)"),
                                     u.create_stretch(), spacing=8)

        is_sequence_row = u.create_row(u.create_label(text=_("Sequence")),
                                       u.create_combo_box(items=["No", "Yes"],
                                                          current_index="@binding(is_sequence_model.value)"),
                                       u.create_stretch(), spacing=8)

        not_sequence_column = u.create_column(u.create_row(u.create_label(text=_("No Sequence Axis"), enabled=False)))

        sequence_column = u.create_column(
            u.create_row(u.create_label(text=_("Sequence Length")),
                         u.create_line_edit(text="@binding(sequence_size_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        is_sequence_stack = u.create_stack(not_sequence_column, sequence_column,
                                           current_index="@binding(is_sequence_model.value)")

        collection_rank_row = u.create_row(u.create_label(text=_("Collection Rank")),
                                           u.create_combo_box(items=["0 (none)", "1 (line)", "2 (scan)"],
                                                              current_index="@binding(collection_rank_model.value)"),
                                           u.create_stretch(), spacing=8)

        no_scan_column = u.create_column(u.create_row(u.create_label(text=_("No Collection Axes"), enabled=False)))

        line_column = u.create_column(
            u.create_row(u.create_label(text=_("Line Length")),
                         u.create_line_edit(text="@binding(line_size_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        scan_column = u.create_column(
            u.create_row(u.create_label(text=_("Scan Width")),
                         u.create_line_edit(text="@binding(scan_width_model.value, converter=int_converter)"),
                         u.create_label(text=_("Scan Height")),
                         u.create_line_edit(text="@binding(scan_height_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        collection_stack = u.create_stack(no_scan_column, line_column, scan_column,
                                          current_index="@binding(collection_rank_model.value)")

        datum_rank_row = u.create_row(u.create_label(text=_("Datum Rank")),
                                      u.create_combo_box(items=["1 (spectrum)", "2 (image)", "1+1 (array of 1d)"],
                                                         current_index="@binding(datum_rank_model.value)"),
                                      u.create_stretch(), spacing=8)

        spectrum_column = u.create_column(
            u.create_row(u.create_label(text=_("Length")),
                         u.create_line_edit(text="@binding(spectrum_size_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        image_column = u.create_column(
            u.create_row(u.create_label(text=_("Width")),
                         u.create_line_edit(text="@binding(image_width_model.value, converter=int_converter)"),
                         u.create_label(text=_("Height")),
                         u.create_line_edit(text="@binding(image_height_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        array_column = u.create_column(
            u.create_row(u.create_label(text=_("Width")),
                         u.create_line_edit(text="@binding(array_width_model.value, converter=int_converter)"),
                         u.create_label(text=_("Height")),
                         u.create_line_edit(text="@binding(array_height_model.value, converter=int_converter)"),
                         u.create_stretch(), spacing=8),
        )

        datum_stack = u.create_stack(spectrum_column, image_column, array_column,
                                     current_index="@binding(datum_rank_model.value)")

        button_row = u.create_row(u.create_stretch(),
                                  u.create_push_button(text=_("Cancel"), on_clicked="close_window"),
                                  u.create_push_button(text=_("Generate"), on_clicked="generate"), spacing=8)

        main_page = u.create_column(title_row, data_type_row,
                                    is_sequence_row, is_sequence_stack,
                                    collection_rank_row, collection_stack,
                                    datum_rank_row, datum_stack,
                                    u.create_spacing(26), button_row, min_width=320 - 24)

        window = u.create_window(main_page, title=_("Generate Data"), margin=12, window_style="tool")

        self.run(window, parent_window=document_controller, persistent_id="generate_data")

    def generate(self, widget: typing.Optional[Declarative.UIWidget] = None) -> None:
        data_shape: typing.Tuple[int, ...] = tuple()
        calibrations: typing.List[Calibration.Calibration] = list()
        is_sequence = False
        collection_rank = 0
        datum_rank = 1

        if self.is_sequence_model.value == 1:
            is_sequence = True
            sequence_size = self.sequence_size_model.value or 1
            data_shape = data_shape + (sequence_size,)
            calibrations.append(Calibration.Calibration(units="s"))

        if self.collection_rank_model.value == 1:
            collection_rank = 1
            line_size = self.line_size_model.value or 1
            data_shape = data_shape + (line_size,)
            calibrations.append(Calibration.Calibration(units="nm"))
        elif self.collection_rank_model.value == 2:
            collection_rank = 2
            scan_height = self.scan_height_model.value or 1
            scan_width = self.scan_width_model.value or 1
            data_shape = data_shape + (scan_height, scan_width)
            calibrations.append(Calibration.Calibration(units="nm"))
            calibrations.append(Calibration.Calibration(units="nm"))

        if self.datum_rank_model.value == 0:
            datum_rank = 1
            spectrum_size = self.spectrum_size_model.value or 1
            data_shape = data_shape + (spectrum_size,)
            calibrations.append(Calibration.Calibration(units="eV"))
        elif self.datum_rank_model.value == 1:
            datum_rank = 2
            image_height = self.image_height_model.value or 1
            image_width = self.image_width_model.value or 1
            data_shape = data_shape + (image_height, image_width)
            calibrations.append(Calibration.Calibration(units="nm"))
            calibrations.append(Calibration.Calibration(units="nm"))
        elif self.datum_rank_model.value == 2:
            collection_rank += 1
            datum_rank = 1
            array_height = self.array_height_model.value or 1
            array_width = self.array_width_model.value or 1
            data_shape = data_shape + (array_height, array_width)
            calibrations.append(Calibration.Calibration(units="nm"))
            calibrations.append(Calibration.Calibration(units="eV"))

        document_model = self.__document_controller.document_model

        data_item = DataItem.DataItem()
        title_value = self.title_model.value
        if title_value:
            data_item.title = title_value
        document_model.append_data_item(data_item)

        data_descriptor = DataAndMetadata.DataDescriptor(is_sequence, collection_rank, datum_rank)

        data_item.reserve_data(data_shape=data_shape, data_dtype=numpy.dtype(numpy.float32), data_descriptor=data_descriptor)

        with data_item.data_ref() as dr:
            data = dr.data
            if data is not None:
                if datum_rank == 1:
                    n = scipy.stats.norm()
                    length = data.shape[-1]
                    data[..., :] = n.pdf(numpy.linspace(n.ppf(1.0 / length), n.ppf(1.0 - 1.0 / length), length))

                # noise
                rng = numpy.random.default_rng()
                data[...] += rng.standard_normal(data.shape)

            dr.data_updated()

        data_item.set_dimensional_calibrations(calibrations)

        display_item = document_model.get_display_item_for_data_item(data_item)
        assert display_item

        self.__document_controller.show_display_item(display_item)



from __future__ import annotations

# standard libraries
import collections
import copy
import datetime
import functools
import gettext
import typing
import weakref

# local libraries
from nion.data import DataAndMetadata
from nion.swift.model import Persistence
from nion.utils import Event
from nion.utils import Observable
from nion.utils.ReferenceCounting import weak_partial
from nion.utils import Registry
from nion.utils import Stream

_ = gettext.gettext

if typing.TYPE_CHECKING:
    from nion.swift.model import DisplayItem


class DynamicString:
    def __init__(self) -> None:
        self.string_stream = Stream.ValueStream[str]()
        self.persistence_stream = Stream.ValueStream[Persistence.PersistentDictType]()

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DynamicString:
        raise NotImplementedError()

    # called by the owner of the dynamic string to connect the slugs to the observed item.
    def connect_item(self, item: typing.Optional[Observable.Observable]) -> None:
        pass


class DynamicStringSlug(Stream.ValueStream[str]):
    """A dynamic string slug is a stream of strings that updates dynamically by observing metadata.

    Subclasses should observe their targets and update the str value by setting the value property in
    response to changes.

    Subclasses can utilize the _item_stream and _display_item_stream properties when configuring
    observation.
    """
    def __init__(self, slug_id: str, default_value: typing.Optional[str] = None) -> None:
        super().__init__(default_value or str())
        self.slug_id = slug_id

        # the item stream is the specific item being observed.
        self._item_stream = Stream.ValueStream[Observable.Observable]()

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DynamicStringSlug:
        return DynamicStringSlug(self.slug_id, self.value)

    def write_to_dict(self) -> Persistence.PersistentDictType:
        d = {"type": self.slug_id}
        if self.value:
            d["value"] = self.value
        return d

    # called by the containing dynamic string during setup.
    def set_item(self, item: typing.Optional[Observable.Observable]) -> None:
        self._item_stream.value = item


ItemPropertySlugType = typing.TypeVar("ItemPropertySlugType")


class ItemPropertySlug(DynamicStringSlug, typing.Generic[ItemPropertySlugType]):
    """A dynamic string slug that observes a property on an item and converts it to a string."""
    def __init__(self,
                 property: str,
                 convert_fn: typing.Optional[typing.Callable[[Observable.Observable, typing.Optional[ItemPropertySlugType]], typing.Optional[str]]],
                 slug: str,
                 default_value: typing.Optional[str] = None) -> None:
        super().__init__(slug, default_value)
        self.__property = property
        self.__convert_fn = convert_fn
        self.__item_value_stream = Stream.PropertyChangedEventStream[ItemPropertySlugType](self._item_stream, self.__property)
        self.__item_value_stream_listener = self.__item_value_stream.value_stream.listen(weak_partial(ItemPropertySlug.__item_property_stream_value_changed, self))
        self.__item_property_stream_value_changed(self.__item_value_stream.value)

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DynamicStringSlug:
        return ItemPropertySlug(self.__property, self.__convert_fn, self.slug_id, self.value)

    def __item_property_stream_value_changed(self, value: typing.Optional[ItemPropertySlugType]) -> None:
        def convert_to_str(value: typing.Optional[ItemPropertySlugType]) -> typing.Optional[str]:
            return str(value) if value is not None else None

        if self.__convert_fn:
            item = self._item_stream.value
            if item:
                self.value = self.__convert_fn(item, value)
            else:
                self.value = None
        else:
            self.value = convert_to_str(value)


class DynamicStringSlugString(DynamicString):
    """A dynamic string, comprised of dynamic string slugs, which produces a stream of rendered strings.

    Clients can observe the rendered_string_stream to get the rendered string.
    """

    def __init__(self, dynamic_string_type: str, dynamic_string_slugs: typing.Optional[typing.Sequence[DynamicStringSlug]] = None) -> None:
        super().__init__()
        self.__dynamic_string_type = dynamic_string_type
        self.__dynamic_string_slugs = list(dynamic_string_slugs) if dynamic_string_slugs else list()

        def combine(*vs: typing.Optional[str]) -> str:
            rendered_title = ' '.join([v or str() for v in vs])
            return ' '.join(rendered_title.strip().split())

        self.__stream = Stream.CombineLatestStream(self.__dynamic_string_slugs, combine)
        self.__stream_listener = self.__stream.value_stream.listen(weak_partial(DynamicStringSlugString.__handle_value_changed, self))
        self.__suppress_stream_listener = False
        self.__item: typing.Optional[Observable.Observable] = None

        self.__handle_value_changed(self.__stream.value)

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DynamicString:
        dynamic_string_copy = DynamicStringSlugString(self.__dynamic_string_type)
        dynamic_string_copy.__copy_from(self)
        return dynamic_string_copy

    def __copy_from(self, dynamic_string: DynamicStringSlugString) -> None:
        self.__suppress_stream_listener = True
        try:
            while self.__stream.stream_list:
                self.__stream.remove_stream(0)
                self.__dynamic_string_slugs.pop(0)
            for dynamic_string_slug in dynamic_string.__dynamic_string_slugs:
                dynamic_string_slug_copy = copy.deepcopy(dynamic_string_slug)
                dynamic_string_slug_copy.set_item(self.__item)
                self.__stream.append_stream(dynamic_string_slug_copy)
                self.__dynamic_string_slugs.append(dynamic_string_slug_copy)
        finally:
            self.__suppress_stream_listener = False
        self.__handle_value_changed(self.__stream.value)

    # called by the owner of the dynamic string to connect the slugs to the observed item.
    def connect_item(self, item: typing.Optional[Observable.Observable]) -> None:
        self.__item = item
        for dynamic_string_slug in self.__dynamic_string_slugs:
            dynamic_string_slug.set_item(item)

    @property
    def _dynamic_string_slugs(self) -> typing.Sequence[DynamicStringSlug]:
        return self.__dynamic_string_slugs

    @property
    def _stream(self) -> Stream.AbstractStream[str]:
        return self.__stream

    def __handle_value_changed(self, value: typing.Optional[str]) -> None:
        if not self.__suppress_stream_listener:
            self.string_stream.value = value or str()
            self.persistence_stream.value = {
                "type": self.__dynamic_string_type,
                "slugs": [dynamic_string_slug.write_to_dict() for dynamic_string_slug in self.__dynamic_string_slugs]
            }


# for testing
class _TestDynamicString(DynamicString):
    value = "green"

    def __init__(self) -> None:
        super().__init__()
        self.string_stream.value = _TestDynamicString.value
        self.persistence_stream.value = {"type": "_test"}

    def read_from_dict(self, d: Persistence.PersistentDictType) -> None:
        self.string_stream.value = _TestDynamicString.value

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DynamicString:
        return _TestDynamicString()


# for testing
_blocked_dynamic_string_factory: typing.Optional[str] = None


# construct a dynamic string from a dictionary. used internally only for now.
def make_dynamic_string(d: typing.Optional[Persistence.PersistentDictType]) -> typing.Optional[DynamicString]:
    if d:
        for component in Registry.get_components_by_type("dynamic-string-factory"):
            dynamic_string_type = d.get("type", None)
            if dynamic_string_type and dynamic_string_type != _blocked_dynamic_string_factory:
                if dynamic_string := typing.cast(DynamicStringFactory, component).make_dynamic_string(dynamic_string_type, d):
                    return dynamic_string
    return None


class DynamicStringFactory(typing.Protocol):
    """A protocol for factory for making a dynamic string from a template identifier."""
    def make_dynamic_string(self, dynamic_string_type: str, d: Persistence.PersistentDictType) -> typing.Optional[DynamicString]: ...


# used for testing
class _TestDynamicStringFactory(DynamicStringFactory):
    def make_dynamic_string(self, dynamic_string_type: str, d: Persistence.PersistentDictType) -> typing.Optional[DynamicString]:
        if dynamic_string_type == "_test":
            dynamic_string = _TestDynamicString()
            dynamic_string.read_from_dict(d)
            return dynamic_string
        if dynamic_string_type == "_slug_test":

            def convert_metadata_to_slug_test_str(item: Observable.Observable, value: typing.Optional[Persistence.PersistentDictType]) -> typing.Optional[str]:
                if value is not None:
                    metadata = value
                    return typing.cast(typing.Optional[str], metadata.get("_slug_test", None))
                else:
                    return None

            return DynamicStringSlugString(dynamic_string_type, [ItemPropertySlug("metadata", convert_metadata_to_slug_test_str, "slug_test")])
        return None


# register the factories.
Registry.register_component(_TestDynamicStringFactory(), {"dynamic-string-factory"})



"""
    Contains classes related to thumbnail generation.
"""

from __future__ import annotations

# standard libraries
import concurrent.futures
import functools
import threading
import typing
import uuid
import weakref

# third-party libraries
import numpy
import numpy.typing

# local libraries
from nion.swift import DisplayPanel
from nion.swift.model import Utility
from nion.swift.model import DisplayItem
from nion.ui import DrawingContext
from nion.ui import UserInterface
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import ReferenceCounting

_NDArray = numpy.typing.NDArray[typing.Any]


class ThumbnailSource:
    """Produce a thumbnail for a display."""
    _executor = concurrent.futures.ThreadPoolExecutor()

    def __init__(self, ui: UserInterface.UserInterface, display_item: DisplayItem.DisplayItem, will_close_fn: typing.Callable[[uuid.UUID], None]) -> None:
        super().__init__()
        self._ui = ui
        self._display_item = display_item
        self.__will_close_fn = will_close_fn

        self.width = 256
        self.height = 256

        self.thumbnail_updated_event = Event.Event()

        self.__display_item = display_item
        self.__recompute_lock = threading.RLock()
        self.__recompute_future: typing.Optional[concurrent.futures.Future[typing.Any]] = None
        # the cache is used to store the thumbnail data persistently. for performance, it is ideal
        # to minimize calling it and instead use the cached value in this class.
        self.__cache = self.__display_item._display_cache
        self.__cache_property_name = "thumbnail_data"
        self.__cache_properties_known = False
        self.__cache_thumbnail_data: typing.Optional[_NDArray] = None
        self.__cache_is_dirty = False

        self.thumbnail_dirty_event = Event.Event()  # for testing

        self.__display_changed_event_listener = display_item.display_changed_event.listen(ReferenceCounting.weak_partial(ThumbnailSource.__thumbnail_changed, self))
        self.__graphics_changed_event_listener = display_item.graphics_changed_event.listen(ReferenceCounting.weak_partial(ThumbnailSource.__graphics_changed, self))

        # initial recompute, if required
        self.__recompute_on_thread()

        self.__display_will_close_listener = display_item.display_item_will_close_event.listen(ReferenceCounting.weak_partial(ThumbnailSource.__display_item_will_close, self))

    def __read_cache_properties(self) -> None:
        if not self.__cache_properties_known:
            self.__cache_thumbnail_data = typing.cast(typing.Optional[_NDArray], self.__cache.get_cached_value(self.__display_item, self.__cache_property_name)) if self.__display_item else None
            self.__cache_is_dirty = self.__cache.is_cached_value_dirty(self.__display_item, self.__cache_property_name) if self.__display_item else False
            self.__cache_properties_known = True
            self.thumbnail_updated_event.fire()

    def __thumbnail_changed(self) -> None:
        self.__cache.set_cached_value_dirty(self.__display_item, self.__cache_property_name)
        self.thumbnail_dirty_event.fire()
        self.__cache_is_dirty = True
        self.__cache_properties_known = True
        self.__recompute_on_thread()

    def __recompute_on_thread(self) -> None:
        with self.__recompute_lock:
            if not self.__recompute_future or self.__recompute_future.done():
                self.__recompute_future = self._executor.submit(self.__recompute_data_if_needed)

    def __graphics_changed(self, graphic_selection: DisplayItem.GraphicSelection) -> None:
        self.__thumbnail_changed()

    def __display_item_will_close(self) -> None:
        # the display item is closing, so these messages should not be triggered, but just in case...
        self.__display_item_about_to_close_listener = typing.cast(typing.Any, None)
        self.__display_changed_event_listener = typing.cast(typing.Any, None)
        self.__graphics_changed_event_listener = typing.cast(typing.Any, None)
        # shut down the thread, if any. avoid deadlock.
        # note: the __display_item still has to be valid to shut down the thread, in case it is still running.
        # clear the display item after shutting down the thread.
        recompute_future: typing.Optional[concurrent.futures.Future[typing.Any]] = None
        with self.__recompute_lock:
            if self.__recompute_future and not self.__recompute_future.done():
                self.__recompute_future.cancel()
                recompute_future = self.__recompute_future
        if recompute_future:
            try:
                concurrent.futures.wait([recompute_future], timeout=10.0)
            except concurrent.futures.CancelledError:
                pass
        self.__will_close_fn(self.__display_item.uuid)
        self.__will_close_fn = typing.cast(typing.Any, None)  # break the reference cycle for faster garbage collection
        self.__display_item = typing.cast(typing.Any, None)

    @property
    def thumbnail_data(self) -> typing.Optional[_NDArray]:
        return self.__cache_thumbnail_data

    def __recompute_data_if_needed(self) -> None:
        self.__read_cache_properties()
        if self._is_thumbnail_dirty:
            self.recompute_data()

    def recompute_data(self) -> None:
        """Compute the data associated with this processor.

        This method is thread safe and may take a long time to return. It should not be called from
         the UI thread. Upon return, the results will be calculated with the latest data available
         and the cache will not be marked dirty.
        """
        ui = self._ui
        try:
            display_item = self.__display_item
            if display_item.display_data_shape and len(display_item.display_data_shape) == 2:
                pixel_shape = Geometry.IntSize(height=512, width=512)
            else:
                pixel_shape = Geometry.IntSize(height=308, width=512)
            drawing_context = DisplayPanel.preview(DisplayPanel.DisplayPanelUISettings(ui), display_item, pixel_shape)
            thumbnail_drawing_context = DrawingContext.DrawingContext()
            thumbnail_drawing_context.scale(self.width / 512, self.height / 512)
            thumbnail_drawing_context.translate(0, (pixel_shape.width - pixel_shape.height) * 0.5)
            thumbnail_drawing_context.add(drawing_context)
            calculated_data = ui.create_rgba_image(thumbnail_drawing_context, self.width, self.height)
        except Exception as e:
            import traceback
            traceback.print_exc()
            traceback.print_stack()
            raise
        if calculated_data is None:
            calculated_data = numpy.zeros((self.height, self.width), dtype=numpy.uint32)
        with self.__recompute_lock:
            self.__cache_thumbnail_data = calculated_data
            self.__cache_is_dirty = False
            self.__cache_properties_known = True
            self.__cache.set_cached_value(self.__display_item, self.__cache_property_name, calculated_data)
        self.thumbnail_updated_event.fire()

    @property
    def _is_thumbnail_dirty(self) -> bool:
        return self.__cache_is_dirty

    @property
    def _is_valid(self) -> bool:
        return self.__display_item is not None


class ThumbnailManager(metaclass=Utility.Singleton):
    """Manages thumbnail sources for displays."""

    def __init__(self) -> None:
        self.__thumbnail_sources: typing.Dict[uuid.UUID, ThumbnailSource] = dict()
        self.__lock = threading.RLock()

    def reset(self) -> None:
        with self.__lock:
            self.__thumbnail_sources.clear()

    def thumbnail_source_for_display_item(self, ui: UserInterface.UserInterface, display_item: DisplayItem.DisplayItem) -> ThumbnailSource:
        """Returned ThumbnailSource must be closed."""
        with self.__lock:
            thumbnail_source = self.__thumbnail_sources.get(display_item.uuid)
            if not thumbnail_source:
                def will_close_fn(display_item_uuid: uuid.UUID) -> None:
                    with self.__lock:
                        del self.__thumbnail_sources[display_item_uuid]

                thumbnail_source = ThumbnailSource(ui, display_item, will_close_fn)
                self.__thumbnail_sources[display_item.uuid] = thumbnail_source
            else:
                assert thumbnail_source._ui == ui
            return thumbnail_source

    def thumbnail_data_for_display_item(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> typing.Optional[_NDArray]:
        with self.__lock:
            thumbnail_source = self.__thumbnail_sources.get(display_item.uuid) if display_item else None
            if thumbnail_source:
                return thumbnail_source.thumbnail_data
            return None



from __future__ import annotations

# standard libraries
import asyncio
import collections
import contextlib
import dataclasses
import datetime
import functools
import logging
import os
import pathlib
import re
import sys
import threading
import time
import traceback
import types

# third party libraries
import typing

import numpy

# local libraries
from nion.utils import DateTime


# datetimes are _local_ datetimes and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
# time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
# daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
# timezone is for conversion and is the Olson timezone string.
# datetime_item is a dictionary with entries for the local_datetime, tz (timezone offset), and
# dst (daylight savings time offset). it may optionally include tz_name (timezone name), if available.
def get_datetime_item_from_datetime(datetime_local: datetime.datetime, tz_minutes: typing.Optional[int] = None,
                                    dst_minutes: typing.Optional[int] = None, timezone: typing.Optional[str] = None) -> typing.Dict[str, typing.Any]:
    # dst is information, tz already includes dst
    datetime_item = dict()
    datetime_item["local_datetime"] = datetime_local.isoformat()
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds())) // 60
    if dst_minutes is None:
        dst_minutes = 60 if time.localtime().tm_isdst else 0
    datetime_item["tz"] = '{0:+03d}{1:02d}'.format(tz_minutes // 60, tz_minutes % 60)
    datetime_item["dst"] = "+{0:02d}".format(dst_minutes)
    if timezone is not None:
        datetime_item["timezone"] = timezone
    return datetime_item


def get_current_datetime_item() -> typing.Dict[str, typing.Any]:
    return get_datetime_item_from_datetime(datetime.datetime.now())


def get_datetime_item_from_utc_datetime(datetime_utc: datetime.datetime, tz_minutes: typing.Optional[int] = None,
                                        dst_minutes: typing.Optional[int] = None,
                                        timezone: typing.Optional[str] = None) -> typing.Dict[str, typing.Any]:
    # dst is information, tz already includes dst
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds())) // 60
    return get_datetime_item_from_datetime(datetime_utc + datetime.timedelta(minutes=tz_minutes), tz_minutes,
                                           dst_minutes, timezone)


local_utcoffset_override: typing.Optional[typing.List[int]] = None  # for testing
def local_utcoffset_minutes(datetime_local: typing.Optional[datetime.datetime] = None) -> int:
    if local_utcoffset_override is not None:
        return local_utcoffset_override[0]
    now = datetime.datetime.now()
    return int((now - now.astimezone(tz=datetime.timezone.utc).replace(tzinfo=None)).total_seconds() / 60)


class TimezoneMinutesToStringConverter:
    def convert(self, value: typing.Optional[int]) -> typing.Optional[str]:
        return "{0:+03d}{1:02d}".format(value // 60, value % 60) if value is not None else None

    def convert_back(self, value: typing.Optional[str]) -> typing.Optional[int]:
        return (int(value[1:3]) * 60 + int(value[3:5])) * (-1 if value[0] == '-' else 1) if value is not None else None


local_timezone_override: typing.Optional[typing.List[str]] = None  # for testing
try:
    import tzlocal

    def get_local_timezone() -> typing.Optional[str]:
        if local_timezone_override is None:
            # see note https://github.com/regebro/tzlocal/issues/117#issuecomment-939351032
            return str(tzlocal.get_localzone())
        else:
            return local_timezone_override[0]
except ImportError:
    def get_local_timezone() -> typing.Optional[str]:
        return None if local_timezone_override is None else local_timezone_override[0]


# return python datetime object from a datetime_item. may return None if the datetime element is
# not properly formatted.
def get_datetime_from_datetime_item(datetime_item: typing.Dict[str, typing.Any]) -> typing.Optional[datetime.datetime]:
    local_datetime = datetime_item.get("local_datetime", str())
    if len(local_datetime) == 26:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S.%f")
    elif len(local_datetime) == 19:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S")
    return None


class Singleton(type):
    def __init__(cls, name: str, bases: typing.Tuple[type], dict: typing.Dict[str, typing.Any]):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args: typing.List[typing.Any], **kw: typing.Dict[str, typing.Any]) -> typing.Any:
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


DirtyValue = typing.Any
CleanValue = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], typing.Tuple[typing.Any], str, float, int, bool, None]


def clean_dict(d0: typing.Dict[str, DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.Dict[str, CleanValue]:
    """Return a json-clean dict. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    d: typing.Dict[str, CleanValue] = dict()
    for key in d0:
        cleaned_item = clean_item_fn(d0[key])
        if cleaned_item is not None:
            d[key] = cleaned_item
    return d


def clean_list(l0: typing.List[DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.List[CleanValue]:
    """Return a json-clean list. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l: typing.List[CleanValue] = list()
    for index, item in enumerate(l0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return l


def clean_tuple(t0: typing.Tuple[DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.Tuple[CleanValue]:
    """Return a json-clean tuple. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l: typing.List[CleanValue] = list()
    for index, item in enumerate(t0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return typing.cast(typing.Tuple[CleanValue], tuple(l))


type_lookup: typing.Mapping[typing.Type[typing.Any], typing.Callable[[DirtyValue], CleanValue]] = {
    dict: clean_dict,
    list: clean_list,
    tuple: clean_tuple,
    numpy.float32: lambda x: float(x),
    numpy.float64: lambda x: float(x),
    numpy.int16: lambda x: int(x),
    numpy.uint16: lambda x: int(x),
    numpy.int32: lambda x: int(x),
    numpy.uint32: lambda x: int(x),
    numpy.int64: lambda x: int(x),
    numpy.uint64: lambda x: int(x),
    float: lambda x: float(x),
    str: lambda x: str(x),
    int: lambda x: int(x),
    numpy.bool_: lambda x: bool(x),
    bool: lambda x: bool(x),
    type(None): lambda x: None
}


type_lookup_no_list: typing.Mapping[typing.Type[typing.Any], typing.Callable[[DirtyValue], CleanValue]] = {
    dict: clean_dict,
    list: clean_tuple,
    tuple: clean_tuple,
    numpy.float32: lambda x: float(x),
    numpy.float64: lambda x: float(x),
    numpy.int16: lambda x: int(x),
    numpy.uint16: lambda x: int(x),
    numpy.int32: lambda x: int(x),
    numpy.uint32: lambda x: int(x),
    numpy.int64: lambda x: int(x),
    numpy.uint64: lambda x: int(x),
    float: lambda x: float(x),
    str: lambda x: str(x),
    int: lambda x: int(x),
    numpy.bool_: lambda x: bool(x),
    bool: lambda x: bool(x),
    type(None): lambda x: None
}


def clean_item(i: DirtyValue) -> CleanValue:
    """Return a json-clean item or None. Will log info message for failure."""
    itype = type(i)
    c = type_lookup.get(itype, None)
    if c:
        return c(i)
    else:
        logging.info("[1] Unable to handle type %s", itype)
        import traceback
        traceback.print_stack()
        return None


def clean_item_no_list(i: DirtyValue) -> CleanValue:
    """Return a json-clean item or None. Will log info message for failure."""
    itype = type(i)
    c = type_lookup_no_list.get(itype, None)
    if c:
        return c(i)
    else:
        logging.info("[1] Unable to handle type %s", itype)
        import traceback
        traceback.print_stack()
        return None


def deep_compare_items(item1: CleanValue, item2: CleanValue) -> bool:
    """Deep compare dict/list/value. Return True if equal."""
    item1_type = type(item1)
    if item1_type != type(item2):
        return False
    if item1_type in (dict,):
        item1_dict = typing.cast(typing.Dict[str, CleanValue], item1)
        item2_dict = typing.cast(typing.Dict[str, CleanValue], item2)
        if item1_dict.keys() != item2_dict.keys():
            return False
        for key in item1_dict.keys():
            if not deep_compare_items(item1_dict[key], item2_dict[key]):
                return False
        return True
    elif item1_type in (list, tuple):
        item1_list = typing.cast(typing.List[CleanValue], item1)
        item2_list = typing.cast(typing.List[CleanValue], item2)
        if len(item1_list) != len(item2_list):
            return False
        for child_item1, child_item2 in zip(item1_list, item2_list):
            if not deep_compare_items(child_item1, child_item2):
                return False
        return True
    return item1 == item2


def parse_version(version: str, count: int = 3, max_count: typing.Optional[int] = None) -> typing.List[int]:
    max_count = max_count if max_count is not None else count
    version_components = [int(version_component) for version_component in version.split(".")]
    assert len(version_components) <= max_count
    while len(version_components) < count:
        version_components.append(0)
    return version_components


def compare_versions(version1: str, version2: str) -> int:
    if version1.startswith("~"):
        version1 = version1[1:]
        version_components1 = parse_version(version1, 1, 3)
        assert len(version_components1) > 1
    elif version1 == "1":  # same as "~1.0"
        version1 = "1.0"
        version_components1 = parse_version(version1, 2, 3)
    else:
        version_components1 = parse_version(version1)
    version_components2 = parse_version(version2)
    # print(version_components1, version_components2)
    for version_component1, version_component2 in zip(version_components1, version_components2):
        if version_component1 > version_component2:
            return 1
        elif version_component1 < version_component2:
            return -1
    return 0


class AtomicFileWriter:
    # see https://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
    # see https://github.com/mahmoud/boltons/blob/1885ed64006982cdd08a70c2cba193e1fe2eb693/boltons/fileutils.py#L343

    def __init__(self, filepath: pathlib.Path):
        self.__filepath = filepath
        self.__temp_filepath = self.__filepath.with_suffix(".temp")
        self.__fp: typing.Optional[typing.TextIO] = None

    def __enter__(self) -> typing.TextIO:
        self.__fp = self.__temp_filepath.open("w")
        return self.__fp

    def __exit__(self, exception_type: typing.Optional[typing.Type[Exception]], value: typing.Optional[Exception], traceback: typing.Optional[types.TracebackType]) -> None:
        assert self.__fp
        self.__fp.flush()
        os.fsync(self.__fp)
        self.__fp.close()
        if exception_type:
            try:
                os.unlink(self.__temp_filepath)
            except Exception:
                pass
        else:
            try:
                os.replace(self.__temp_filepath, self.__filepath)
                if hasattr(os, "O_DIRECTORY"):
                    # ensure the directory has been updated. not available all the time on Windows.
                    dirfd = os.open(os.path.dirname(self.__filepath), os.O_DIRECTORY)
                    os.fsync(dirfd)
                    os.close(dirfd)
            except OSError:
                try:
                    os.unlink(self.__temp_filepath)
                except Exception:
                    pass
                raise


@dataclasses.dataclass
class FPSRecord:
    window_elapsed_times: typing.List[float] = dataclasses.field(default_factory=list)
    window_elapsed_time: float = 0.0
    last_tick: typing.Optional[float] = None
    last_window_frame_mean: float = 0.0
    last_window_frame_std: float = 0.0


def fps_tick(fps_id: str) -> str:
    v = globals().setdefault("__fps_" + fps_id, FPSRecord())
    next_time = time.perf_counter()
    if v.last_tick is not None:
        elapsed_time = next_time - v.last_tick
        v.window_elapsed_times.append(elapsed_time)
        v.window_elapsed_time += elapsed_time
    if v.window_elapsed_time > 1.0:
        v.last_window_frame_mean = numpy.mean(v.window_elapsed_times)
        v.last_window_frame_std = numpy.std(v.window_elapsed_times)
        v.window_elapsed_time = 0.0
        v.window_elapsed_times = list()
    v.last_tick = next_time
    return fps_get(fps_id)


def fps_get(fps_id: str) -> str:
    v = globals().setdefault("__fps_" + fps_id, FPSRecord())
    m = v.last_window_frame_mean
    r = 1 / m if m != 0.0 else 0.0
    s = v.last_window_frame_std
    return str(round(r * 10) / 10.0) + "fps / " + str(int(m * 1000)) + "ms / " + str(int(s * 1000)) + "ms"


Trace = collections.namedtuple("Trace", ["start_time", "last_time_ref", "min_elapsed", "discard", "all"])


def trace_calls(trace: Trace, frame: types.FrameType, event: str, arg: typing.Any) -> None:
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    assert caller is not None
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    t = time.time()
    last_elapsed = t - trace.last_time_ref[0]
    if not trace.discard or trace.discard not in func_filename:
        if last_elapsed > trace.min_elapsed:
            s = "         " if last_elapsed > 0.001 else ""
            trace.all.append(":%s%12.5f %12.5f Call to %s on line %s of %s from line %s of %s" % (
                s, last_elapsed, t - trace.start_time, func_name, func_line_no, func_filename, caller_line_no, caller_filename))
        trace.last_time_ref[0] = t


def begin_trace(min_elapsed: float, discard: typing.Optional[str]) -> Trace:
    trace = Trace(start_time=time.time(), last_time_ref=[0], min_elapsed=min_elapsed, discard=discard, all=list())
    sys.settrace(typing.cast(typing.Callable[[types.FrameType, str, typing.Any], None], functools.partial(trace_calls, trace)))
    return trace


def end_trace(trace: Trace) -> None:
    sys.settrace(None)
    logging.debug("\n".join(trace.all))
    logging.debug("TOTAL: %s", len(trace.all))


@contextlib.contextmanager
def trace(min_elapsed: float = 0.0, discard: typing.Optional[str] = None) -> typing.Generator[None, None, None]:
    t = begin_trace(min_elapsed, discard)
    yield
    end_trace(t)


def sample_stack_all(count: int = 10, interval: float = 0.1) -> None:
    """Sample the stack in a thread and print it at regular intervals."""

    def print_stack_all(l: threading.RLock, ll: typing.MutableSequence[str]) -> None:
        l1 = list()
        l1.append("*** STACKTRACE - START ***")
        code = []
        for threadId, stack in sys._current_frames().items():
            sub_code = []
            sub_code.append("# ThreadID: %s" % threadId)
            for filename, lineno, name, line in traceback.extract_stack(stack):
                sub_code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    sub_code.append("  %s" % (line.strip()))
            if not "in select" in sub_code[-2] and \
               not "in wait" in sub_code[-2] and \
               not "in _worker" in sub_code[-2] and \
               not "in print_stack_all" in sub_code[-2] and \
               not "in sample_stack_all" in sub_code[-2] and \
               not "in checkcache" in sub_code[-2] and \
               not "do_sleep" in sub_code[-2] and \
               not "sleep" in sub_code[-1] and \
               not any(["in do_sample" in s for s in sub_code]):
                code.extend(sub_code)
        for line in code:
            l1.append(line)
        l1.append("*** STACKTRACE - END ***")
        with l:
            ll.extend(l1)

    def do_sample() -> None:
        l = threading.RLock()
        ll: typing.List[str] = list()
        for i in range(count):
            print_stack_all(l, ll)
            time.sleep(interval)
        with l:
            print("\n".join(ll))

    threading.Thread(target=do_sample).start()


def simplify_filename(filename: str, replacement_char: str = '_', maximum_length: int = 128) -> str:
    # This function replaces any illegal characters in a file name. Not a full path.
    # macOS illegal characters = \0 /
    # linux illegal characters = \0 /
    # Windows illegal characters = < > : / \ | ? * " ASCII values 0-31 (non-printable chars)
    # Windows illegal filenames = CON PRN AUX NUL COM1 COM2 COM3 COM4 COM5 COM6 COM7 COM8 COM9
    #                             LPT1 LPT2 LPT3 LPT4 LPT5 LPT6 LPT7 LPT8 LPT9
    # those files names are illegal with or without a suffix, upper and lower case
    # Windows file names cannot start or end with periods or spaces

    assert maximum_length > 5
    assert len(replacement_char) == 1

    # since windows is the most restrictive and covers the others, just focus on that
    illegal_chars = r'^[. ]|[<>:/\\|?*\"]|[\0-\31]|[. ]$'
    illegal_names = ['CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', 'COM8', 'COM9',
                     'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9']

    # fix path length first so we don't uncover . or space at the start
    updated_filename = filename
    if len(updated_filename) > maximum_length:
        path = pathlib.Path(updated_filename)
        stem = path.stem
        suffix = path.suffix

        if len(suffix) >= maximum_length:
            updated_filename = replacement_char + suffix[1:]
        else:
            updated_filename = stem[:maximum_length - len(suffix)] + suffix

    # replace illegal characters with replacement_char
    updated_filename = re.sub(illegal_chars, replacement_char, updated_filename)

    # if filename (without suffix) is illegal name - prepend replacement_char
    file_name_without_suffix = pathlib.Path(updated_filename).stem
    if file_name_without_suffix.upper() in illegal_names:
        updated_filename = replacement_char + updated_filename
        if len(updated_filename) > maximum_length:
            # should only happen if the illegal name and suffix were already maximum_length
            updated_filename = replacement_char + updated_filename[2:]

    return updated_filename


class TraceCloseable:
    all: typing.List[TraceCloseable] = list()

    def __init__(self) -> None:
        self.__tb = traceback.extract_stack()
        TraceCloseable.all.append(self)

    def close(self) -> None:
        TraceCloseable.all.remove(self)

    @classmethod
    def print_leftovers(cls) -> None:
        import sys
        for x in TraceCloseable.all:
            print("**************************")
            print(f"LEAKED {x}")
            print(f"ALLOCATED HERE")
            for line in traceback.StackSummary.from_list(typing.cast(typing.List[typing.Tuple[str, int, str, typing.Optional[str]]], x.__tb)).format():
                print(line, file=sys.stderr, end="")
            print("^^^^^^^^^^^^^^^^^^^^^^^^^^")


class TestEventLoop:
    def __init__(self, event_loop: typing.Optional[asyncio.AbstractEventLoop] = None):
        logging.disable(logging.CRITICAL)  # suppress new_event_loop debug message
        self.__event_loop = event_loop if event_loop else asyncio.new_event_loop()
        logging.disable(logging.NOTSET)

    def close(self) -> None:
        # give cancelled tasks a chance to finish
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        self.__event_loop.run_until_complete(asyncio.gather(*asyncio.all_tasks(loop=self.__event_loop)))
        # now close
        # due to a bug in Python libraries, the default executor needs to be shutdown explicitly before the event loop
        # see http://bugs.python.org/issue28464
        default_executor = getattr(self.__event_loop, "_default_executor", None)
        if default_executor:
            default_executor.shutdown()
        self.__event_loop.close()
        self.__event_loop = typing.cast(asyncio.AbstractEventLoop, None)

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop


class Timer:
    def __init__(self, *, threshold: float = 0.0) -> None:
        self.start_time_ns = time.perf_counter_ns()
        self.last_time_ns = self.start_time_ns
        self.threshold = threshold

    def mark(self, title: str) -> None:
        current_time = time.perf_counter_ns()
        if self.threshold == 0.0 or (current_time - self.last_time_ns) // 1E9 > self.threshold:
            print(f"{title}: {(current_time - self.start_time_ns) // 1E6}ms +{(current_time - self.last_time_ns) // 1E6}ms")
        self.last_time_ns = current_time
